@@ -63,10 +63,13 @@
 #include <string.h>
 #include <errno.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <sys/wait.h>
 #include <syslog.h>
 #include <time.h>
 #include <unistd.h>
+#include <fcntl.h>
+#include <sys/vt.h>
 
 #include <security/pam_appl.h>
 #include <security/pam_misc.h>
@@ -193,6 +196,119 @@ int change_uid (const struct passwd *info)
 	return 0;
 }
 
+/*
+ * Cleanup ~/.xsession-errors.
+ *
+ * The function looks for .xsession-errors in the current directory, so when it
+ * is called the current directory must be the user's homedir.
+ *
+ * The function also assumes that we are running as the user. As a consequence
+ * it does not worry about symlink attacks, because they would only be possible
+ * if the user's home directory is group or world writable.
+ *
+ * curdirname is the name of the current directory, and it is only used when
+ * logging error messages.
+ *
+ * The function returns true on success, false on failure.
+ */
+int cleanup_xse(off_t maxsize, const char* curdirname)
+{
+	int ret = 0;
+	int xse_fd = -1;
+	struct stat xse_st;
+
+	xse_fd = open(".xsession-errors", O_WRONLY | O_CREAT, 0600);
+	if (xse_fd < 0)
+	{
+		perror ("open ~/.xsession-errors");
+		syslog (LOG_ERR, "cannot open `%s/%s': %m\n", curdirname, ".xsession-errors");
+		goto cleanup;
+	}
+	if (fstat(xse_fd, &xse_st) < 0)
+	{
+		perror ("stat ~/.xsession-errors");
+		syslog (LOG_ERR, "cannot stat `%s/%s': %m\n", curdirname, ".xsession-errors");
+		goto cleanup;
+	}
+	if (xse_st.st_size > maxsize)
+	{
+		if (ftruncate(xse_fd, 0) < 0)
+		{
+			perror ("truncating ~/.xsession-errors");
+			syslog (LOG_ERR, "cannot truncate `%s/%s': %m\n", curdirname, ".xsession-errors");
+			goto cleanup;
+		}
+	}
+
+	/* If we made it so far, we succeeded */
+	ret = 1;
+
+cleanup:
+	if (xse_fd >= 0)
+		close(xse_fd);
+	return ret;
+}
+
+static int try_vtstate(const char* dev, struct vt_stat* vtstat)
+{
+	int res = 0;
+	int fd = open(dev, O_WRONLY | O_NOCTTY, 0);
+	if (fd < 0)
+		goto cleanup;
+	if (ioctl (fd, VT_GETSTATE, vtstat) < 0)
+		goto cleanup;
+	res = 1;
+
+cleanup:
+	if (fd >= 0) close(fd);
+	return res;
+}
+
+static int get_vtstate(struct vt_stat* vtstat)
+{
+	if (try_vtstate("/dev/tty", vtstat)) return 1;
+	if (try_vtstate("/dev/tty0", vtstat)) return 1;
+	if (try_vtstate("/dev/console", vtstat)) return 1;
+	return 0;
+}
+
+/*
+ * Allocate a new vt, open it and return the file descriptor and the vt number.
+ *
+ * Searching the vt starts at the initial value of vtnum.
+ */
+int open_vt(int *vtnum)
+{
+	char vtname[15];
+	int res = -1;
+	struct vt_stat vtstat;
+	unsigned short vtmask;
+
+	if (!get_vtstate(&vtstat))
+	{
+		fprintf (stderr, _("%s: cannot find or open the console\n"), Prog);
+		goto cleanup;
+	}
+
+	for (vtmask = 1 << *vtnum; vtstat.v_state & vtmask; ++*vtnum, vtmask <<= 1)
+		;
+	if (!vtmask) {
+		fprintf (stderr, _("%s: all VTs seem to be busy\n"), Prog);
+		goto cleanup;
+	}
+
+	snprintf(vtname, 15, "/dev/tty%d", *vtnum);
+
+	res = open(vtname, O_RDWR | O_NOCTTY, 0);
+	if (res < 0) {
+		fprintf (stderr, _("%s: cannot open %s: %m\n"), Prog, vtname);
+		goto cleanup;
+	}
+
+cleanup:
+	return res;
+}
+
 /* Signal handler for parent process later */
 static void catch_signals (int sig)
 {
@@ -404,8 +520,11 @@ static int nodm_monitor(int argc, char **argv)
 	const char* opt_session = NODM_SESSION;
 	char xinit[BUFSIZ];
 	char xoptions[BUFSIZ];
+	char xoptions1[BUFSIZ];
 	char* cp;
 	int mst;
+	int vt_fd = -1;
+	int vt_num;
 
 	/* Parse command line options */
 	while (1)
@@ -443,14 +562,32 @@ static int nodm_monitor(int argc, char **argv)
 
 	syslog(LOG_INFO, "Starting nodm monitor");
 
+	{
+		char startvt[BUFSIZ];
+		string_from_env(startvt, "NODM_FIRST_VT", "7");
+		vt_num = strtoul(startvt, NULL, 10);
+	}
+	if ((vt_fd = open_vt(&vt_num)) == -1)
+	{
+		fprintf (stderr, _("%s: cannot allocate a virtual terminal\n"), Prog);
+		return 1;
+	}
+
 	/* Read the configuration from the environment */
 	cp = getenv("NODM_MIN_SESSION_TIME");
 	mst = cp ? atoi(cp) : 60;
 	string_from_env(xinit, "NODM_XINIT", "/usr/bin/xinit");
 	string_from_env(xoptions, "NODM_X_OPTIONS", "");
 
+	if (xoptions[0] == 0)
+		snprintf(xoptions1, BUFSIZ, "vt%d", vt_num);
+	else
+		snprintf(xoptions1, BUFSIZ, "vt%d %s", vt_num, xoptions);
+
 	setenv("NODM_RUN_SESSION", "1", 1);
-	run_and_restart(xinit, opt_session, xoptions[0] == 0 ? NULL : xoptions, mst);
+	run_and_restart(xinit, opt_session, xoptions1, mst);
+
+	close(vt_fd);
 
 	return 0;
 }
@@ -643,7 +780,9 @@ static int nodm_session(int argc, char **argv)
 	unsetenv("NODM_MIN_SESSION_TIME");
 	unsetenv("NODM_RUN_SESSION");
 
-	chdir (pwent.pw_dir);
+	if (chdir (pwent.pw_dir) == 0)
+		/* Truncate ~/.xsession-errors */
+		cleanup_xse(0, pwent.pw_dir);
 
 	args[0] = "/bin/sh";
 	args[1] = "-l";
